#!/usr/bin/env python3 
# SPDX-License-Identifier: MIT
# src/pdflinkcheck/gui.py
import tkinter as tk
from tkinter import filedialog, ttk, messagebox, PhotoImage
import sys
from pathlib import Path
from typing import Optional
import unicodedata
from importlib.resources import files
import pyhabitat
import ctypes

# --- Core Imports ---
from pdflinkcheck.report import run_report_and_call_exports
from pdflinkcheck.version_info import get_version_from_pyproject
from pdflinkcheck.io import get_first_pdf_in_cwd, get_friendly_path, PDFLINKCHECK_HOME
from pdflinkcheck.environment import pymupdf_is_available, clear_all_caches, is_in_git_repo

class RedirectText:
    """A class to redirect sys.stdout messages to a Tkinter Text widget."""
    def __init__(self, text_widget):
        self.text_widget = text_widget

    def write(self, string):
        self.text_widget.insert(tk.END, string)
        self.text_widget.see(tk.END) 
        self.text_widget.update_idletasks() 

    def flush(self, *args):
        pass

class PDFLinkCheckerApp(tk.Tk):

    # --- Theme & Visual Initialization ---

    def _initialize_forest_theme(self):
        theme_dir = files("pdflinkcheck.data.themes.forest") 
        self.tk.call("source", str(theme_dir / "forest-light.tcl")) 
        self.tk.call("source", str(theme_dir / "forest-dark.tcl")) 

    def _toggle_theme(self):
        if ttk.Style().theme_use() == "forest-light":
            ttk.Style().theme_use("forest-dark")
        elif ttk.Style().theme_use() == "forest-dark":
            ttk.Style().theme_use("forest-light")

    def _set_icon(self):
        icon_dir = files("pdflinkcheck.data.icons") 
        try:
            png_path = icon_dir.joinpath("Logo-150x150.png")
            if png_path.exists():
                self.icon_img = PhotoImage(file=str(png_path))
                self.iconphoto(True, self.icon_img)
        except Exception:
            pass
        try:
            icon_path = icon_dir.joinpath("red_pdf_512px.ico")
            if icon_path.exists():
                self.iconbitmap(str(icon_path))
        except Exception:
            pass
            
    # --- Lifecycle & Initialization ---

    def __init__(self):
        super().__init__()

        self._initialize_forest_theme()
        ttk.Style().theme_use("forest-dark")

        title_suffix = ""
        self.title(f"PDF Link Check v{get_version_from_pyproject()}{title_suffix}")
        #self.geometry("900x650") # Slightly wider for new layout columns
        self.geometry("800x600")

        self._set_icon()

        # --- 1. Variable State Management ---
        self.pdf_path = tk.StringVar(value="")
        self.pdf_library_var = tk.StringVar(value="PyMuPDF")
        self.do_export_report_json_var = tk.BooleanVar(value=True) 
        self.do_export_report_txt_var = tk.BooleanVar(value=True) 
        self.current_report_text = None
        self.current_report_data = None
        
        # Track exported file paths for direct opening
        self.last_json_path: Optional[Path] = None
        self.last_txt_path: Optional[Path] = None

        if not pymupdf_is_available():
            self.pdf_library_var.set("pypdf")

        # --- 2. Widget Construction ---
        self._create_widgets()
        self._initialize_menubar()

    def _initialize_menubar(self):
        """Builds the application menu bar."""
        menubar = tk.Menu(self)
        self.config(menu=menubar)

        tools_menu = tk.Menu(menubar, tearoff=0)
        menubar.add_cascade(label="Tools", menu=tools_menu)

        tools_menu.add_command(label="Toggle Theme", command=self._toggle_theme)
        tools_menu.add_command(label="Clear Output Window", command=self._clear_output_window)
        tools_menu.add_command(label="Copy Output to Clipboard", command=self._copy_output_to_clipboard)
        tools_menu.add_command(label="Clear Cache", command=self._clear_all_caches)

        tools_menu.add_separator()
        tools_menu.add_command(label="License", command=self._show_license)
        tools_menu.add_command(label="Readme", command=self._show_readme)
        tools_menu.add_command(label="I Have Questions", command=self._show_i_have_questions)

    # --- UI Component Building ---

    def _create_widgets(self):
        """Primary layout definition using a grid-based control panel and pack-based output."""
        
        # --- Control Frame (Top) ---
        control_frame = ttk.Frame(self, padding="10")
        control_frame.pack(fill='x')

        # === Row 0: File Selection ===
        file_selection_frame = ttk.Frame(control_frame)
        file_selection_frame.grid(row=0, column=0, columnspan=3, padx=0, pady=5, sticky='ew')
        
        ttk.Label(file_selection_frame, text="PDF Path:").pack(side=tk.LEFT, padx=(0, 5))
        ttk.Entry(file_selection_frame, textvariable=self.pdf_path, width=50).pack(side=tk.LEFT, fill='x', expand=True, padx=5)
        ttk.Button(file_selection_frame, text="Browse...", command=self._select_pdf).pack(side=tk.LEFT, padx=(5, 5))
        ttk.Button(file_selection_frame, text="Copy Path", command=self._copy_pdf_path).pack(side=tk.LEFT, padx=(0, 0))

        # === Row 1: Configuration & Export Jumps ===
        
        # 1.1 PDF Library Group
        pdf_library_frame = ttk.LabelFrame(control_frame, text="Backend Engine:")
        pdf_library_frame.grid(row=1, column=0, padx=5, pady=5, sticky='nsew')

        ttk.Radiobutton(pdf_library_frame, text="PyMuPDF", variable=self.pdf_library_var, value="PyMuPDF").pack(side='left', padx=5, pady=1) 
        ttk.Radiobutton(pdf_library_frame, text="pypdf", variable=self.pdf_library_var, value="pypdf").pack(side='left', padx=5, pady=1)

        # 1.2 Export Format Selection
        export_config_frame = ttk.LabelFrame(control_frame, text="Export Enabled:")
        export_config_frame.grid(row=1, column=1, padx=5, pady=5, sticky='nsew')
        
        ttk.Checkbutton(export_config_frame, text="JSON", variable=self.do_export_report_json_var).pack(side=tk.LEFT, padx=10)
        ttk.Checkbutton(export_config_frame, text="TXT", variable=self.do_export_report_txt_var).pack(side=tk.LEFT, padx=10)

        # 1.3 Export File Actions
        self.export_actions_frame = ttk.LabelFrame(control_frame, text="Open Report Files:")
        self.export_actions_frame.grid(row=1, column=2, padx=5, pady=5, sticky='nsew')

        self.btn_open_json = ttk.Button(self.export_actions_frame, text="Open JSON", command=lambda: self._open_export_file("json"))
        self.btn_open_json.pack(side=tk.LEFT, padx=5, pady=2)

        self.btn_open_txt = ttk.Button(self.export_actions_frame, text="Open TXT", command=lambda: self._open_export_file("txt"))
        self.btn_open_txt.pack(side=tk.LEFT, padx=5, pady=2)

        # === Row 3: Action Buttons ===
        run_analysis_btn = ttk.Button(control_frame, text="▶ Run Analysis", command=self._run_report_gui, style='Accent.TButton')
        run_analysis_btn.grid(row=3, column=0, columnspan=2, pady=10, sticky='ew', padx=(0, 5))

        clear_window_btn = ttk.Button(control_frame, text="Clear Output Window", command=self._clear_output_window)
        clear_window_btn.grid(row=3, column=2, pady=10, sticky='ew', padx=(5, 0))

        # Grid configuration for even distribution
        control_frame.grid_columnconfigure(0, weight=1)
        control_frame.grid_columnconfigure(1, weight=1)
        control_frame.grid_columnconfigure(2, weight=1)

        # --- Output Frame (Bottom) ---
        output_frame = ttk.Frame(self, padding=(10, 2, 10, 10))
        output_frame.pack(fill='both', expand=True)

        output_header_frame = ttk.Frame(output_frame)
        output_header_frame.pack(fill='x', pady=(0, 5))
        
        ttk.Label(output_header_frame, text="Analysis Report Logs:").pack(side=tk.LEFT, fill='x', expand=True)

        ttk.Button(output_header_frame, text="▼ Bottom", command=self._scroll_to_bottom, width=8).pack(side=tk.RIGHT, padx=(0, 5)) 
        ttk.Button(output_header_frame, text="▲ Top", command=self._scroll_to_top, width=6).pack(side=tk.RIGHT, padx=(5, 5))

        # Scrollable Text Area
        text_scroll_frame = ttk.Frame(output_frame)
        text_scroll_frame.pack(fill='both', expand=True, padx=5, pady=5)
        
        self.output_text = tk.Text(text_scroll_frame, wrap=tk.WORD, state=tk.DISABLED, bg='#2b2b2b', fg='#ffffff', font=('Monospace', 10))
        self.output_text.pack(side=tk.LEFT, fill='both', expand=True)

        scrollbar = ttk.Scrollbar(text_scroll_frame, command=self.output_text.yview)
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.output_text['yscrollcommand'] = scrollbar.set

    # --- Event Handlers & Business Logic ---


    def _select_pdf(self):
        if self.pdf_path.get():
            initialdir = str(Path(self.pdf_path.get()).parent)
        
        # MSIX environments often start in System32; redirect to Home for better UX.
        elif pyhabitat.is_msix(): 
            initialdir = str(Path.home())

        # Ideal for CLI usage and portable usage
        else: 
            initialdir = str(Path.cwd())

        file_path = filedialog.askopenfilename(
            initialdir=initialdir,
            defaultextension=".pdf",
            filetypes=[("PDF files", "*.pdf"), ("All files", "*.*")]
        )
        if file_path:
            self.pdf_path.set(get_friendly_path(file_path))

    def _copy_pdf_path(self):
        path_to_copy = self.pdf_path.get()
        if path_to_copy:
            try:
                self.clipboard_clear()
                self.clipboard_append(path_to_copy)
                messagebox.showinfo("Copied", "PDF Path copied to clipboard.")
            except tk.TclError as e:
                messagebox.showerror("Copy Error", f"Clipboard access blocked: {e}")
        else:
            messagebox.showwarning("Copy Failed", "PDF Path field is empty.")

    def _run_report_gui(self):
        """Executes the analysis and updates export button states."""
        pdf_path_str = self._assess_pdf_path_str()
        if not pdf_path_str:
            return

        export_format = ""
        if self.do_export_report_json_var.get():
            export_format += "JSON"
        if self.do_export_report_txt_var.get():
            export_format += "TXT" 

        pdf_library = self.pdf_library_var.get().lower()

        # Prep output window
        self.output_text.config(state=tk.NORMAL)
        self.output_text.delete('1.0', tk.END)

        original_stdout = sys.stdout
        sys.stdout = RedirectText(self.output_text)
        
        try:
            report_results = run_report_and_call_exports(
                pdf_path=pdf_path_str,
                export_format=export_format,
                pdf_library=pdf_library, 
            )
            self.current_report_text = report_results.get("text", "")
            self.current_report_data = report_results.get("data", {})
            
            # Capture file paths from the report results
            self.last_json_path = report_results.get("files", {}).get("export_path_json")
            self.last_txt_path = report_results.get("files", {}).get("export_path_txt")
            
        except Exception as e:
            messagebox.showinfo(
                "Engine Fallback",
                f"Error encountered with {pdf_library}: {e}\n\nFalling back to pypdf."
            )
            self.pdf_library_var.set("pypdf")
        finally:
            sys.stdout = original_stdout
            self.output_text.config(state=tk.DISABLED)
 
    def _open_export_file(self, file_type: str):
        """Launches the system default editor using pyhabitat's robust ladder logic."""
        target_path = self.last_json_path if file_type == "json" else self.last_txt_path
        
        if target_path and Path(target_path).exists():
            try:
                # pyhabitat 1.1.1+ automatically detects if it should be non-blocking
                # based on the absence of a TTY/REPL (perfect for this Tkinter GUI).
                pyhabitat.edit_textfile(target_path)
            except Exception as e:
                messagebox.showerror("Open Error", f"Failed to open {file_type.upper()} report: {e}")
        else:
            messagebox.showwarning("File Not Found", f"The {file_type.upper()} report file does not exist. \n\nPlease click 'Run Analysis' to generate one.")

    def _assess_pdf_path_str(self):
        pdf_path_str = self.pdf_path.get().strip()
        if not pdf_path_str: 
            pdf_path_str = get_first_pdf_in_cwd()
            if not pdf_path_str:
                self._display_error("No PDF found in current directory.")
                return

        p = Path(pdf_path_str).expanduser().resolve()
        if not p.exists():
            self._display_error(f"PDF file not found at: {p}")
            return
            
        return str(p)

    # --- Utility Methods ---

    def _clear_output_window(self):
        self.output_text.config(state=tk.NORMAL)
        self.output_text.delete('1.0', tk.END)
        self.output_text.config(state=tk.DISABLED)

    def _copy_output_to_clipboard(self):
        content = self.output_text.get('1.0', tk.END)
        self.clipboard_clear()
        self.clipboard_append(content)
        messagebox.showinfo("Clipboard", "Output buffer copied to clipboard.")

    def _scroll_to_top(self):
        self.output_text.see('1.0')

    def _scroll_to_bottom(self):
        self.output_text.see(tk.END)

    def _clear_all_caches(self):
        clear_all_caches()
        messagebox.showinfo("Caches Cleared", f"All caches have been cleared.\nPyMuPDF available: {pymupdf_is_available()}")

    def _display_error(self, message):
        self.output_text.config(state=tk.NORMAL)
        self.output_text.insert(tk.END, f"[ERROR] {message}\n", 'error')
        self.output_text.tag_config('error', foreground='red')
        self.output_text.see(tk.END)
        self.output_text.config(state=tk.DISABLED)

    # --- Modal Documentation Windows ---

    def _show_license(self):
        self._display_resource_window("LICENSE", "Software License")

    def _show_readme(self):
        self._display_resource_window("README.md", "pdflinkcheck README.md")

    def _show_i_have_questions(self):
        self._display_resource_window("I Have Questions.md", "I Have Questions.md")

    def _display_resource_window(self, filename: str, title: str):
        """Generic modal window for displaying data directory text files, with dev-mode fallback."""
        content = None
        try:
            # Primary: Try to read from embedded package resources
            content = (files("pdflinkcheck.data") / filename).read_text(encoding="utf-8")
        except FileNotFoundError:
            if is_in_git_repo():
                # Development mode: embedded files not present → copy from root
                messagebox.showinfo(
                    "Development Mode",
                    f"Embedded {filename} not found.\nTrying to copy from project root..."
                )
                try:
                    from pdflinkcheck.datacopy import ensure_data_files_for_build
                    ensure_data_files_for_build()  # This should copy LICENSE, README.md, etc.
                    
                    # Retry reading after copy
                    content = (files("pdflinkcheck.data") / filename).read_text(encoding="utf-8")
                except ImportError:
                    messagebox.showerror(
                        "Fallback Failed",
                        "Cannot import datacopy module. Please ensure pdflinkcheck.datacopy exists."
                    )
                    return
                except Exception as e:
                    messagebox.showerror(
                        "Copy Failed",
                        f"Failed to copy {filename} from root: {e}"
                    )
                    return
            else:
                # Packaged mode: no fallback possible
                messagebox.showerror(
                    "Resource Missing",
                    f"Embedded file '{filename}' not found.\n"
                    "This indicates a packaging or installation issue."
                )
                return
        except Exception as e:
            messagebox.showerror("Read Error", f"Failed to read {filename}: {e}")
            return

        # Sanitize content for Tkinter display
        content = sanitize_glyphs_for_tkinter(content)

        # Display in modal window
        win = tk.Toplevel(self)
        win.title(title)
        win.geometry("700x500")

        txt = tk.Text(win, wrap=tk.WORD, font=('Monospace', 10), padx=10, pady=10)
        txt.insert(tk.END, content)
        txt.config(state=tk.DISABLED)

        sb = ttk.Scrollbar(win, command=txt.yview)
        txt['yscrollcommand'] = sb.set

        sb.pack(side=tk.RIGHT, fill=tk.Y)
        txt.pack(fill='both', expand=True)

        win.transient(self)
        win.grab_set()
    
    def _display_resource_window_defunct(self, filename: str, title: str):
        """Generic modal window for displaying data directory text files."""
        try:
            content = (files("pdflinkcheck.data") / filename).read_text(encoding="utf-8")
            content = sanitize_glyphs_for_tkinter(content)
        except Exception:
            messagebox.showerror("Error", f"Could not load {filename}")
            return

        win = tk.Toplevel(self)
        win.title(title)
        win.geometry("700x500")
        
        txt = tk.Text(win, wrap=tk.WORD, font=('Monospace', 10), padx=10, pady=10)
        txt.insert(tk.END, content)
        txt.config(state=tk.DISABLED)
        
        sb = ttk.Scrollbar(win, command=txt.yview)
        txt['yscrollcommand'] = sb.set
        
        sb.pack(side=tk.RIGHT, fill=tk.Y)
        txt.pack(fill='both', expand=True)
        
        win.transient(self)
        win.grab_set()

# --- Helper Functions ---

def sanitize_glyphs_for_tkinter(text: str) -> str:
    """Removes non-ASCII characters to prevent rendering errors in Tkinter."""
    normalized = unicodedata.normalize('NFKD', text)
    sanitized = normalized.encode('ascii', 'ignore').decode('utf-8')
    return sanitized.replace('  ', ' ')

def start_gui(time_auto_close: int = 0):
    """Entry point for launching the GUI application."""
    print("pdflinkcheck: start_gui ...")
    app = PDFLinkCheckerApp()

    # Window Focus Management
    app.lift() 
    app.wm_attributes("-topmost", True) 
    app.after(200, lambda: app.wm_attributes("-topmost", False))
    app.focus_force()

    if pyhabitat.on_windows():
        hwnd = app.winfo_id() 
        ctypes.windll.user32.SetForegroundWindow(hwnd)
    
    if time_auto_close > 0:
        app.after(time_auto_close, app.destroy)

    app.mainloop()
    print("pdflinkcheck: gui closed.")

if __name__ == "__main__":
    start_gui()



#!/usr/bin/env python3 
# SPDX-License-Identifier: MIT
# src/pdflinkcheck/gui.py
import tkinter as tk
from tkinter import filedialog, ttk, messagebox, PhotoImage
import sys
from pathlib import Path
from typing import Optional # Added Optional
import unicodedata
from importlib.resources import files
import pyhabitat
import ctypes

# --- Core Imports ---
from pdflinkcheck.report import run_report_and_call_exports
from pdflinkcheck.version_info import get_version_from_pyproject
from pdflinkcheck.io import get_first_pdf_in_cwd, get_friendly_path, PDFLINKCHECK_HOME
from pdflinkcheck.environment import pymupdf_is_available, clear_all_caches, is_in_git_repo

class RedirectText:
    """A class to redirect sys.stdout messages to a Tkinter Text widget."""
    def __init__(self, text_widget):
        self.text_widget = text_widget

    def write(self, string):
        """Insert the incoming string into the Text widget."""
        self.text_widget.insert(tk.END, string)
        self.text_widget.see(tk.END) # Scroll to the end
        self.text_widget.update_idletasks() # Refresh GUI to allow real timie updates << If suppress: The mainloop will handle updates efficiently without forcing them, , but info appears outdated when a new file is analyzed. Immediate feedback is better.

    def flush(self, *args):
        """Required for file-like objects, but does nothing here."""
        pass


class PDFLinkCheckerApp(tk.Tk):

    # --- Theme & Visual Initialization --

    def _initialize_forest_theme(self):
        from importlib.resources import files 
        # Path to pdflinkcheck/data/themes/forest/ 
        theme_dir = files("pdflinkcheck.data.themes.forest") 
        # Load the theme files
        self.tk.call("source", str(theme_dir / f"forest-light.tcl")) 
        self.tk.call("source", str(theme_dir / f"forest-dark.tcl")) 

    def _toggle_theme(self):
        # You could instead assign the dark to light of a single theme here
        """
        Calls light/dark toggle for the forest theme with self._toggle_theme_just_forest()
        """
        return self._toggle_theme_forest()

    def _toggle_theme_forest(self):
        if ttk.Style().theme_use() == "forest-light":
            ttk.Style().theme_use("forest-dark")
        elif ttk.Style().theme_use() == "forest-dark":
            ttk.Style().theme_use("forest-light")

    def _set_icon(self):
        from importlib.resources import files 
    
        # Path to pdflinkcheck/data/icons/
        icon_dir = files("pdflinkcheck.data.icons") 
        try:
            # 1. For Linux/macOS (and modern Windows), use iconphoto with a PNG
            png_path = icon_dir.joinpath("Logo-150x150.png")
            if png_path.exists():
                self.icon_img = PhotoImage(file=str(png_path))
                self.iconphoto(True, self.icon_img)
        except:
            pass
        try:
            # 2. Specifically for Windows taskbar/window title.
            # We wrap this in a try because it will fail on Linux.
            # If both the PNG and the ICO succeed, the ICO will override the PNG. 
            icon_path = icon_dir.joinpath("red_pdf_512px.ico")
            if icon_path.exists():
                self.iconbitmap(str(icon_path))
        except:
            pass
    
    # --- Lifecycle & Initialization ---

    def __init__(self):
        super().__init__()

    
        self._initialize_forest_theme() # load but do not set internally
        
        ttk.Style().theme_use("forest-dark") # but if you use _toggle_theme_just_forest(), then you had better do this

        if is_in_git_repo() and not pyhabitat.as_pyinstaller() and not pyhabitat.is_pyz():
            # Checking for PYZ is overkill, because a PYZ s not expected to carry a .git directory, which is a check that is already completed.
            title_suffix = ""# " [Development]"
        else:
            title_suffix = ""
        
        self.title(f"PDF Link Check v{get_version_from_pyproject()}{title_suffix}")
        self.geometry("800x600")

        self._set_icon()

        
        # --- 1. Initialize Variables ---
        self.pdf_path = tk.StringVar(value="")
        self.pdf_library_var = tk.StringVar(value="PyMuPDF")
        self.do_export_report_json_var = tk.BooleanVar(value=True) 
        self.do_export_report_txt_var = tk.BooleanVar(value=True) 
        self.current_report_text = None
        self.current_report_data = None

        # Track exported file paths for direct opening
        self.last_json_path: Optional[Path] = None
        self.last_txt_path: Optional[Path] = None

        if not pymupdf_is_available():
            print(f"pymupdf_is_available: {pymupdf_is_available()}")
            self.pdf_library_var.set("pypdf")

        # --- 2. Create Widgets ---
        self._create_widgets()
        #self._initialize_menubar()
        
        # --- 3. Set Initial Dependent Widget States ---
        self._toggle_json_export()
        self._toggle_txt_export()

        # --- Menubar with dropdown ---
        menubar = tk.Menu(self)
        self.config(menu=menubar)

        # Tools menu
        tools_menu = tk.Menu(menubar, tearoff=0)
        menubar.add_cascade(label="Tools", menu=tools_menu)

        tools_menu.add_command(label="Toggle Theme", command=self._toggle_theme)
        tools_menu.add_command(label="Clear Cache", command=self._clear_all_caches)

        # Add existing License/Readme to tools menu
        tools_menu.add_separator()
        tools_menu.add_command(label="License", command=self._show_license)
        tools_menu.add_command(label="Readme", command=self._show_readme)
        tools_menu.add_command(label="I Have Questions", command=self._show_i_have_questions)
    # In class PDFLinkCheckerApp:

    def _copy_pdf_path(self):
        """Copies the current PDF path from the Entry widget to the system clipboard."""
        path_to_copy = self.pdf_path.get()
        
        if path_to_copy:
            try:
                # Clear the clipboard
                self.clipboard_clear()
                # Append the path string to the clipboard
                self.clipboard_append(path_to_copy)
                # Notify the user (optional, but good UX)
                messagebox.showinfo("Copied", "PDF Path copied to clipboard.")
            except tk.TclError as e:
                # Handle cases where clipboard access might be blocked
                messagebox.showerror("Copy Error", f"Failed to access the system clipboard: {e}")
        else:
            messagebox.showwarning("Copy Failed", "The PDF Path field is empty.")
    
    def _scroll_to_top(self):
        """Scrolls the output text widget to the top."""
        self.output_text.see('1.0') # '1.0' is the index for the very first character

    def _scroll_to_bottom(self):
        """Scrolls the output text widget to the bottom."""
        self.output_text.see(tk.END) # tk.END is the index for the position just after the last character

    def _clear_all_caches(self):
        """Clear caches and show confirmation."""
        clear_all_caches()
        messagebox.showinfo("Caches Cleared", f"All caches have been cleared.\nPyMuPDF available: {pymupdf_is_available()}")

    def _show_license(self):
        """
        Reads the embedded LICENSE file (AGPLv3) and displays its content in a new modal window.
        """
        try:
            # Use the Traversable object's read_text() method.
            # This handles files located inside zip archives (.pyz, pipx venvs) correctly.
            license_path_traversable = files("pdflinkcheck.data") / "LICENSE"
            license_content = license_path_traversable.read_text(encoding="utf-8")
            
        except FileNotFoundError:
            if is_in_git_repo():
                messagebox.showinfo(
                    "Local Development Mode",
                    "Embedded data files not found – copying from root..."
                )
                try:
                    from pdflinkcheck.datacopy import ensure_package_license, ensure_data_files_for_build
                    #ensure_package_license()
                    ensure_data_files_for_build()  
                    # Retry display
                    self._show_license() 
                    return
                except Exception as e:
                    messagebox.showerror("Copy Failed", f"Could not copy file: {e}")
            else:
                messagebox.showerror(
                    "Packaging Error",
                    "Embedded file not found. This indicates a problem with the package build/installation."
                )
            return
        
        except Exception as e:
            messagebox.showerror("Read Error", f"Failed to read embedded LICENSE file: {e}")
            return

        # --- Display in a New Toplevel Window ---
        license_window = tk.Toplevel(self)
        license_window.title("Software License")
        license_window.geometry("600x400")
        
        # Text widget for content
        text_widget = tk.Text(license_window, wrap=tk.WORD, font=('Monospace', 10), padx=10, pady=10)
        text_widget.insert(tk.END, license_content)
        text_widget.config(state=tk.DISABLED)
        
        # Scrollbar
        scrollbar = ttk.Scrollbar(license_window, command=text_widget.yview)
        text_widget['yscrollcommand'] = scrollbar.set
        
        # Layout
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        text_widget.pack(fill='both', expand=True)
        
        # Make the window modal (optional, but good practice for notices)
        license_window.transient(self)
        license_window.grab_set()
        self.wait_window(license_window)

    def _show_readme(self):
        """
        Reads the embedded README.md file and displays its content in a new modal window.
        """
        try:
            # Use the Traversable object's read_text() method.
            # This handles files located inside zip archives (.pyz, pipx venvs) correctly.
            readme_path_traversable = files("pdflinkcheck.data") / "README.md"
            readme_content = readme_path_traversable.read_text(encoding="utf-8")
            readme_content = sanitize_glyphs_for_tkinter(readme_content)
        except FileNotFoundError:
            if is_in_git_repo():
                messagebox.showinfo(
                    "Local Development Mode",
                    "Embedded data files not found – copying from root..."
                )
                try:
                    from pdflinkcheck.datacopy import ensure_package_readme, ensure_data_files_for_build 
                    #ensure_package_readme()
                    ensure_data_files_for_build()  
                    # Retry display
                    self._show_readme() 
                    return
                except Exception as e:
                    messagebox.showerror("Copy Failed", f"Could not copy file: {e}")
            else:
                messagebox.showerror(
                    "Packaging Error",
                    "Embedded file not found. This indicates a problem with the package build/installation."
                )
            return
        except Exception as e:
            messagebox.showerror("Read Error", f"Failed to read embedded README.md file: {e}")
            return

        # --- Display in a New Toplevel Window ---
        readme_window = tk.Toplevel(self)
        readme_window.title("pdflinkcheck README.md")
        readme_window.geometry("600x400")
        
        # Text widget for content
        text_widget = tk.Text(readme_window, wrap=tk.WORD, font=('Monospace', 10), padx=10, pady=10)
        text_widget.insert(tk.END, readme_content)
        text_widget.config(state=tk.DISABLED)
        
        # Scrollbar
        scrollbar = ttk.Scrollbar(readme_window, command=text_widget.yview)
        text_widget['yscrollcommand'] = scrollbar.set
        
        # Layout
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        text_widget.pack(fill='both', expand=True)
        
        # Make the window modal (optional, but good practice for notices)
        readme_window.transient(self)
        readme_window.grab_set()
        self.wait_window(readme_window)

    def _show_i_have_questions(self):
        """
        Reads the embedded I Have Questions.md file and displays its content in a new modal window.
        """
        try:
            # Use the Traversable object's read_text() method.
            # This handles files located inside zip archives (.pyz, pipx venvs) correctly.
            i_have_questions_path_traversable = files("pdflinkcheck.data") / "I Have Questions.md"
            i_have_questions_content = i_have_questions_path_traversable.read_text(encoding="utf-8")
            i_have_questions_content = sanitize_glyphs_for_tkinter(i_have_questions_content)
        except FileNotFoundError:
            messagebox.showerror("Read Error", f"Failed to read embedded 'I Have Questions.md' file.")
            return

        # --- Display in a New Toplevel Window ---
        i_have_questions_window = tk.Toplevel(self)
        i_have_questions_window.title("I Have Questions.md")
        i_have_questions_window.geometry("600x400")
        
        # Text widget for content
        text_widget = tk.Text(i_have_questions_window, wrap=tk.WORD, font=('Monospace', 10), padx=10, pady=10)
        text_widget.insert(tk.END, i_have_questions_content)
        text_widget.config(state=tk.DISABLED)
        
        # Scrollbar
        scrollbar = ttk.Scrollbar(i_have_questions_window, command=text_widget.yview)
        text_widget['yscrollcommand'] = scrollbar.set
        
        # Layout
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        text_widget.pack(fill='both', expand=True)
        
        # Make the window modal (optional, but good practice for notices)
        i_have_questions_window.transient(self)
        i_have_questions_window.grab_set()
        self.wait_window(i_have_questions_window)

    def _create_widgets(self):
        # --- Control Frame (Top) ---
        control_frame = ttk.Frame(self, padding="10")
        control_frame.pack(fill='x')

        # Row 0: File Selection

        # === File Selection Frame (Row 0) ===
        file_selection_frame = ttk.Frame(control_frame)
        file_selection_frame.grid(row=0, column=0, columnspan=3, padx=0, pady=5, sticky='ew')
        
        # Elements are now packed/gridded within file_selection_frame
        
        # Label
        ttk.Label(file_selection_frame, text="PDF Path:").pack(side=tk.LEFT, padx=(0, 5))
        
        # Entry (Path Display)
        ttk.Entry(file_selection_frame, textvariable=self.pdf_path, width=50).pack(side=tk.LEFT, fill='x', expand=True, padx=5)
        # The Entry field (column 1) must expand horizontally within its frame
        # Since we are using PACK for this frame, we use fill='x', expand=True on the Entry.
        
        # Browse Button
        ttk.Button(file_selection_frame, text="Browse...", command=self._select_pdf).pack(side=tk.LEFT, padx=(5, 5))

        # Copy Button
        # NOTE: Removed leading spaces from " Copy Path"
        ttk.Button(file_selection_frame, text="Copy Path", command=self._copy_pdf_path).pack(side=tk.LEFT, padx=(0, 0))
        
        # === END: File Selection Frame ===

        # --- Report brevity options ----
        #report_brevity_frame.grid(row=1, column=0, padx=5, pady=5, sticky='nsew')


        # --- PDF Library Selection ---
        # Create a labeled group for the PDF options
        pdf_library_frame = ttk.LabelFrame(control_frame, text="Select PDF Library:")
        pdf_library_frame.grid(row=1, column=1, padx=5, pady=5, sticky='nsew')

        # Radio options inside the frame
        ttk.Radiobutton(
            pdf_library_frame,
            text="PyMuPDF",
            variable=self.pdf_library_var,
            value="PyMuPDF",
            
        ).pack(side='left', padx=5, pady=1)   

        ttk.Radiobutton(
            pdf_library_frame,
            text="pypdf",
            variable=self.pdf_library_var,
            value="pypdf",
        ).pack(side='left', padx=5, pady=1)

        export_group_frame = ttk.LabelFrame(control_frame, text="Export Format:")
        #export_group_frame = ttk.LabelFrame(control_frame, text = "Export Filetype Selection:")
        export_group_frame.grid(row=1, column=2, padx=5, pady=5, sticky='nseew') # Placed in the original Checkbutton's column
        
        ttk.Checkbutton(
            export_group_frame, 
            #text="Export Report",
            text = "JSON" ,
            variable=self.do_export_report_json_var
        ).pack(side=tk.LEFT, padx=(0, 5)) # Pack Checkbutton to the left with small internal padding
        ttk.Checkbutton(
            export_group_frame, 
            text = "TXT" ,
            #state=tk.DISABLED,
            variable=self.do_export_report_txt_var,
        ).pack(side=tk.LEFT, padx=(0, 5)) # Pack Checkbutton to the left with small internal padding
        
        # Row 3: Run Button, Export Filetype selection, License Button, and readme button
        # 1. Run Button (Spans columns 0 and 1)
        run_analysis_btn = ttk.Button(control_frame, text="▶ Run Analysis", command=self._run_report_gui, style='Accent.TButton')
        run_analysis_btn.grid(row=3, column=0, columnspan=2, pady=10, sticky='ew', padx=(0, 5))

        """
        # 2. Create a Frame to hold the two file link buttons (This frame goes into column 2)
        info_btn_frame = ttk.Frame(control_frame)
        info_btn_frame.grid(row=3, column=2, columnspan=1, pady=10, sticky='ew', padx=(5, 0))
        # Ensure the info button frame expands to fill its column
        info_btn_frame.grid_columnconfigure(0, weight=1)
        info_btn_frame.grid_columnconfigure(1, weight=1)

        # 3. Placeholder buttons inside the info button frame
        info_1_btn = ttk.Button(info_btn_frame, text="Empty1", command=self._do_stuff_1)
        # Use PACK or a 2-column GRID inside the info_btn_frame. GRID is cleaner here.
        info_1_btn.grid(row=0, column=0, sticky='ew', padx=(0, 2)) # Left side of the frame

        info_2_btn = ttk.Button(info_btn_frame, text="Empty2", command=self._do_stuff_2)
        info_2_btn.grid(row=0, column=1, sticky='ew', padx=(2, 0)) # Right side of the frame
        """

        # Force the columns to distribute space evenly
        control_frame.grid_columnconfigure(0, weight=2)
        control_frame.grid_columnconfigure(1, weight=1)
        control_frame.grid_columnconfigure(2, weight=1)

        # --- Output Frame (Bottom) ---
        output_frame = ttk.Frame(self, padding=(10, 2, 10, 10)) # Left, Top, Right, Bottom
        output_frame.pack(fill='both', expand=True)

        output_header_frame = ttk.Frame(output_frame)
        output_header_frame.pack(fill='x', pady=(0, 5))
        
        # Label
        ttk.Label(output_header_frame, text="Analysis Report Output:").pack(side=tk.LEFT, fill='x', expand=True)

        # Scroll to Bottom Button # put this first so that it on the right when the Top button is added on the left.
        bottom_btn = ttk.Button(output_header_frame, text="▼ Bottom", command=self._scroll_to_bottom, width=8)
        bottom_btn.pack(side=tk.RIGHT, padx=(0, 5)) 

        # Scroll to Top Button
        top_btn = ttk.Button(output_header_frame, text="▲ Top", command=self._scroll_to_top, width=6)
        top_btn.pack(side=tk.RIGHT, padx=(5, 5))

        # Open Report Button
        self.open_report_btn = ttk.Button(output_header_frame, text="Open Report", command=self._open_report_text)
        self.open_report_btn.pack(side=tk.RIGHT, padx=(5, 5))
        
        
        # ----------------------------------------------------
        
        
        # Scrollable Text Widget for output
        # Use an internal frame for text and scrollbar to ensure correct packing
        text_scroll_frame = ttk.Frame(output_frame)
        text_scroll_frame.pack(fill='both', expand=True, padx=5, pady=5)
        
        self.output_text = tk.Text(text_scroll_frame, wrap=tk.WORD, state=tk.DISABLED, bg='#333333', fg='white', font=('Monospace', 10))
        self.output_text.pack(side=tk.LEFT, fill='both', expand=True) # Text fills and expands

        # Scrollbar (Scrollbar must be packed AFTER the text widget)
        scrollbar = ttk.Scrollbar(text_scroll_frame, command=self.output_text.yview)
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.output_text['yscrollcommand'] = scrollbar.set # Link text widget back to scrollbar

    def _select_pdf(self):
        if self.pdf_path.get():
            initialdir = str(Path(self.pdf_path.get()).parent)
        elif pyhabitat.is_msix(): 
            # Don't look in system 32; add additonal checks for any expected installed GUI-only rollouts, to various stores. 
            # CLI should default to cwd(), whether installed or portable.
            # awaiting pyhabitat 1.0.54
            initialdir = str(Path.home())
        else: # best for CLI usage and portable usage
            initialdir = str(Path.cwd())

        file_path = filedialog.askopenfilename(
            initialdir=initialdir,
            defaultextension=".pdf",
            filetypes=[("PDF files", "*.pdf"), ("All files", "*.*")]
        )
        if file_path:
            self.pdf_path.set(get_friendly_path(file_path))
    
    def _toggle_json_export(self):
        """Checkbox toggle for json filetype report."""
        if self.do_export_report_json_var.get():
            pass # placeholder # no side effects

    def _toggle_txt_export(self):
        """Checkbox toggle for TXT filetype report."""
        if self.do_export_report_txt_var.get():
            pass # placeholder # no side effects
    


    def _toggle_pdf_library(self):
        selected_lib = self.pdf_library_var.get().lower()
        try:
            self.pdf_library_var.set("pypdf" if selected_lib == "pymupdf" else "pymupdf")
        except Exception:
            pass

    def _assess_pdf_path_str(self):
        pdf_path_str = self.pdf_path.get().strip()
        if not pdf_path_str: 
            pdf_path_str = get_first_pdf_in_cwd()
            if not pdf_path_str:
                self._display_error("Error: No PDF found in current directory.")
                return

        p = Path(pdf_path_str).expanduser().resolve()

        if not p.exists():
            self._display_error(f"Error: PDF file not found at: {p}")
            return
            
        # Use the resolved string version for the rest of the function
        pdf_path_str_assessed = str(p)
        return pdf_path_str_assessed

    def _run_report_gui(self):
        
        pdf_path_str = self._assess_pdf_path_str()
        if not pdf_path_str:
            return

        export_format = None # default value, if selection is not made (if selection is not active)
        export_format = ""
        if self.do_export_report_json_var.get():
            export_format = export_format + "JSON"
        if self.do_export_report_txt_var.get():
            export_format = export_format + "TXT" 

        pdf_library = self._discern_pdf_library()

        
        # 1. Clear previous output and enable editing
        self.output_text.config(state=tk.NORMAL)
        self.output_text.delete('1.0', tk.END)

        # 2. Redirect standard output to the Text widget
        original_stdout = sys.stdout
        sys.stdout = RedirectText(self.output_text)
        
        try:
            # 3. Call the core logic function
            #self.output_text.insert(tk.END, "--- Starting Analysis ---\n")
            report_results = run_report_and_call_exports(
                pdf_path=pdf_path_str,
                export_format=export_format,
                pdf_library = pdf_library, 
            )
            self.current_report_text = report_results.get("text", "")
            self.current_report_data = report_results.get("data", {})

            #self.output_text.insert(tk.END, "\n--- Analysis Complete ---\n")

        except Exception as e:
            # Inform the user in the GUI with a clean message
            messagebox.showinfo(  # Changed from showwarning – less alarming
                "PyMuPDF Not Available",
                "PyMuPDF is not installed or not working.\n"
                "Switched to pypdf engine automatically.\n\n"
                "To use the faster PyMuPDF engine:\n"
                "1. Install it: pip install pymupdf\n"
                "2. Go to Tools → Clear Cache\n"
                "3. Run analysis again"
            )
            self._toggle_pdf_library()
    
            
        finally:
            # 4. Restore standard output and disable editing
            sys.stdout = original_stdout
            self.output_text.config(state=tk.DISABLED)
            # -- This call to _toggle_pdf_library() fails currently --
            
    def _discern_pdf_library(self):
        selected_lib = self.pdf_library_var.get().lower()
        
        if selected_lib == "pymupdf":
            self._display_msg("Using high-speed PyMuPDF engine.")
        elif selected_lib == "pypdf":
            self._display_msg("Using pure-python pypdf engine.")
        return selected_lib
    
    def _display_msg(self, message):
        # Ensure output is in normal state to write
        original_state = self.output_text.cget('state')
        if original_state == tk.DISABLED:
            self.output_text.config(state=tk.NORMAL)
            
        #self.output_text.delete('1.0', tk.END)
        self.output_text.insert(tk.END, f"{message}\n", 'msg')
        self.output_text.tag_config('msg')#, foreground='red')
        self.output_text.see(tk.END)

    def _display_error(self, message):
        # Ensure output is in normal state to write
        original_state = self.output_text.cget('state')
        if original_state == tk.DISABLED:
            self.output_text.config(state=tk.NORMAL)
            
        #self.output_text.delete('1.0', tk.END)
        self.output_text.insert(tk.END, f"[ERROR] {message}\n", 'error')
        self.output_text.tag_config('error', foreground='red')
        self.output_text.see(tk.END)

        # Restore state
        self.output_text.config(state=tk.DISABLED)

    def _open_report_text(self):
        """Opens the LATEST analysis text in an editor, regardless of export settings."""
        # 1. Check our internal buffer, not the window or the disk
        if not self.current_report_text:
            messagebox.showwarning("Open Failed", "No analysis data available. Please run an analysis first.")
            return

        try:
            # 2. Always create a 'viewing' file in a temp directory or .tmp folder
            # This prevents clobbering an actual user-saved report.
            pdf_name = Path(self.pdf_path.get()).stem if self.pdf_path.get() else "report"
            view_path = PDFLINKCHECK_HOME / f"LAST_REPORT_{pdf_name}.txt"
            
            # 3. Write our buffer to this 'View' file
            view_path.write_text(self.current_report_text, encoding="utf-8")
            
            # 4. Open with pyhabitat
            pyhabitat.edit_textfile(view_path)
            
        except Exception as e:
            messagebox.showerror("View Error", f"Could not launch editor: {e}")
    
def sanitize_glyphs_for_tkinter(text: str) -> str:
    """
    Converts complex Unicode characters (like emojis and symbols) 
    into their closest ASCII representation, ignoring those that 
    cannot be mapped. This prevents the 'empty square' issue in Tkinter.
    """
    # 1. Normalize the text (NFKD converts composite characters to their base parts)
    normalized = unicodedata.normalize('NFKD', text)
    
    # 2. Encode to ASCII and decode back. 
    # The 'ignore' flag is crucial: it removes any characters 
    # that don't have an ASCII representation.
    sanitized = normalized.encode('ascii', 'ignore').decode('utf-8')
    
    # 3. Clean up any resulting double spaces or artifacts
    sanitized = sanitized.replace('  ', ' ')
    return sanitized

def auto_close_window(root, delay_ms:int = 0):
    """
    Schedules the Tkinter window to be destroyed after a specified delay.
    """
    if delay_ms > 0:
        print(f"Window is set to automatically close in {delay_ms/1000} seconds.")
        root.after(delay_ms, root.destroy)
    else:
        return



def start_gui(time_auto_close:int=0):
    """
    Entry point function to launch the application.
    """
    print("pdflinkcheck: start_gui ...")
    
    tk_app = PDFLinkCheckerApp()

    # Bring window to front 
    tk_app.lift() 
    #tk_app.attributes('-topmost', True) 
    #tk_app.after(100, lambda: tk_app.attributes('-topmost', False)) 
    tk_app.wm_attributes("-topmost", True) 
    tk_app.after(200, lambda: tk_app.wm_attributes("-topmost", False))
    tk_app.deiconify()
    tk_app.focus_force()

    # Win32 nudge (optional but helpful)
    if pyhabitat.on_windows():
        hwnd = tk_app.winfo_id() 
        ctypes.windll.user32.SetForegroundWindow(hwnd)
    
    # Ths is called in the CLI by the --auto-close flag value, for CI scripted testing purposes (like in .github/workflows/build.yml) 
    auto_close_window(tk_app, time_auto_close)

    tk_app.mainloop()
    print("pdflinkcheck: gui closed.")

if __name__ == "__main__":
    start_gui()



#!/usr/bin/env python3 
# SPDX-License-Identifier: MIT
# pdflinkcheck/analyze_pymupdf.py

import sys
from pathlib import Path
import logging
from typing import Dict, Any, Optional, List

logging.getLogger("fitz").setLevel(logging.ERROR) 

from pdflinkcheck.environment import pymupdf_is_available
try:
    if pymupdf_is_available():
        import fitz  # PyMuPDF
    else:
        fitz = None
except ImportError:
    fitz = None

"""
Inspect target PDF for both URI links and for GoTo links.
"""

# Helper function: Prioritize 'from'
def get_link_rect(link_dict):
    """
    Retrieves the bounding box for the link using the reliable 'from' key
    provided by PyMuPDF's link dictionary.

    Args:
        link_dict: A dictionary representing a single link/annotation 
                   returned by `page.get_links()`.

    Returns:
        A tuple of four floats (x0, y0, x1, y1) representing the 
        rectangular coordinates of the link on the page, or None if the 
        bounding box data is missing.
    """
    # 1. Use the 'from' key, which returns a fitz.Rect object or None
    rect_obj = link_dict.get('from') 
    
    if rect_obj:
        # 2. Extract the coordinates using the standard Rect properties 
        #    (compatible with all recent PyMuPDF versions)
        return (rect_obj.x0, rect_obj.y0, rect_obj.x1, rect_obj.y1)
    
    # 3. Fallback to None if 'from' is missing
    return None

def get_anchor_text(page, link_rect):
    if not link_rect:
        return "N/A: Missing Rect"

    try:
        # 1. Convert to fitz.Rect and normalize
        rect = fitz.Rect(link_rect)
        if rect.is_empty:
            return "N/A: Rect Error"

        # 2. Use asymmetric expansion (similar to the pypdf logic)
        # 10 points horizontal to catch wide characters/kerning
        # 3 points vertical to stay within the line
        search_rect = fitz.Rect(
            rect.x0 - 10, 
            rect.y0 - 3, 
            rect.x1 + 10, 
            rect.y1 + 3
        )

        # 3. Extract all words on the page
        # Each word is: (x0, y0, x1, y1, "text", block_no, line_no, word_no)
        words = page.get_text("words")
        
        anchor_parts = []
        for w in words:
            word_rect = fitz.Rect(w[:4])
            # Check if the word intersects our expanded link rectangle
            if word_rect.intersects(search_rect):
                anchor_parts.append(w[4])

        cleaned_text = " ".join(anchor_parts).strip()
        
        return cleaned_text if cleaned_text else "N/A: No Visible Text"
            
    except Exception:
        return "N/A: Rect Error"
    
def get_anchor_text_stable(page, link_rect):
    """
    Extracts text content using the link's bounding box coordinates.
    The bounding box is slightly expanded to ensure full characters are captured.

    Args:
        page: The fitz.Page object where the link is located.
        link_rect: A tuple of four floats (x0, y0, x1, y1) representing the 
                   link's bounding box.

    Returns:
        The cleaned, extracted text string, or a placeholder message 
        if no text is found or if an error occurs.
    """
    if not link_rect:
        return "N/A: Missing Rect"

    try:
        # 1. Convert the coordinate tuple back to a fitz.Rect object
        rect = fitz.Rect(link_rect)
        
        # --- CRITICAL STEP: Check for invalid/empty rect AFTER conversion ---
        # If the rect is invalid (e.g., width or height is <= 0), skip it
        # Note: fitz.Rect will often auto-normalize, but this explicit check is safer.
        if rect.is_empty or rect.width <= 0 or rect.height <= 0:
            return "N/A: Rect Error (Zero/Negative Dimension)"

        # 2. Expand the rect slightly to capture full characters (1 unit in each direction)
        #    This method avoids the proprietary/unstable 'from_expanded' or 'from_rect' methods.
        expanded_rect = fitz.Rect(
            rect.x0 - 1, 
            rect.y0 - 1, 
            rect.x1 + 1, 
            rect.y1 + 1
        )
        
        # 3. Get the text within the expanded bounding box
        anchor_text = page.get_textbox(expanded_rect)
        
        # 4. Clean up whitespace and non-printing characters
        cleaned_text = " ".join(anchor_text.split())
        
        if cleaned_text:
            return cleaned_text
        else:
            return "N/A: No Visible Text"
            
    except Exception:
        # Fallback for unexpected errors in rect conversion or retrieval
        return "N/A: Rect Error"

def analyze_toc_fitz(doc):
    """
    Extracts the structural Table of Contents (PDF Bookmarks/Outline) 
    from the PDF document using PyMuPDF's built-in functionality.

    Args:
        doc: The open fitz.Document object.

    Returns:
        A list of dictionaries, where each dictionary represents a TOC entry 
        with 'level', 'title', and 'target_page' (1-indexed).
    """
    toc = doc.get_toc()
    toc_data = []
    
    for level, title, page_num in toc:
        # fitz pages are 1-indexed for TOC!
        toc_data.append({
            'level': level,
            'title': title,
            'target_page': page_num
        })
        
    return toc_data


# 2. Updated Main Inspection Function to Include Text Extraction
#def inspect_pdf_hyperlinks_fitz(pdf_path):
def extract_toc_pymupdf(pdf_path):
    """
    Opens a PDF, iterates through all pages and extracts the structural table of contents (TOC/bookmarks).

    Args:
        pdf_path: The file system path (str) to the target PDF document.

    Returns:
        A list of dictionaries representing the structural TOC/bookmarks.
    """
    try:
        doc = fitz.open(pdf_path)
        structural_toc = analyze_toc_fitz(doc)
    except Exception as e:
        print(f"An error occurred: {e}", file=sys.stderr)
    return structural_toc


def serialize_fitz_object(obj):
    """Converts a fitz object (Point, Rect, Matrix) to a serializable type."""
    # Meant to avoid known Point errors like: '[ERROR] An unexpected error occurred during analysis: Report export failed due to an I/O error: Object of type Point is not JSON serializable'
    if obj is None:
        return None
    
    # 1. Handle fitz.Point (has x, y)
    if hasattr(obj, 'x') and hasattr(obj, 'y') and not hasattr(obj, 'x0'):
        return (obj.x, obj.y)
        
    # 2. Handle fitz.Rect and fitz.IRect (has x0, y0)
    if hasattr(obj, 'x0') and hasattr(obj, 'y0'):
        return (obj.x0, obj.y0, obj.x1, obj.y1)
        
    # 3. Handle fitz.Matrix (has a, b, c, d, e, f)
    if hasattr(obj, 'a') and hasattr(obj, 'b') and hasattr(obj, 'c'):
        return (obj.a, obj.b, obj.c, obj.d, obj.e, obj.f)
        
    # 4. Fallback: If it's still not a primitive type, convert it to a string
    if not isinstance(obj, (str, int, float, bool, list, tuple, dict)):
        # Examples: hasattr(value, 'rect') and hasattr(value, 'point'):
        # This handles Rect and Point objects that may slip through
        return str(obj)
        
    # Otherwise, return the object as is (it's already primitive)
    return obj


def extract_links_pymupdf(pdf_path):
    """
    Opens a PDF, iterates through all pages and extracts all link annotations. 
    It categorizes the links into External, Internal, or Other actions, and extracts the anchor text.
    
    Args:
        pdf_path: The file system path (str) to the target PDF document.

    Returns:
        A list of dictionaries, where each dictionary is a comprehensive 
           representation of an active hyperlink found in the PDF.
        
    """
    links_data = []
    try:
        doc = fitz.open(pdf_path)        

        for page_num in range(doc.page_count):
            page = doc.load_page(page_num)
            
            for link in page.get_links():

                page_obj = doc.load_page(page_num)
                link_rect = get_link_rect(link)
                
                rect_obj = link.get("from")
                xref = link.get("xref")
                #print(f"rect_obj = {rect_obj}")
                #print(f"xref = {xref}")
                

                # --- Examples of various keys associated with various link instances ---
                #print(f"keys: list(link) = {list(link)}")
                # keys: list(link) = ['kind', 'xref', 'from', 'page', 'viewrect', 'id']
                # keys: list(link) = ['kind', 'xref', 'from', 'uri', 'id']
                # keys: list(link) = ['kind', 'xref', 'from', 'page', 'view', 'id']

                # 1. Extract the anchor text
                anchor_text = get_anchor_text(page_obj, link_rect)

                # 2. Extract the target and kind
                target = ""
                kind = link.get('kind')
                
                
                link_dict = {
                    'page': int(page_num) + 1, # accurate for link location, add 1
                    'rect': link_rect,
                    'link_text': anchor_text,
                    'xref':xref
                }
                
                # A. Clean Geom. Objects: Use the helper function on 'to' / 'destination'
                # Use the clean serialize_fitz_object() helper function on all keys that might contain objects
                destination_view = serialize_fitz_object(link.get('to'))

                # B. Correct Internal Link Page Numbering (The -1 correction hack)
                # This will be skipped by URI, which is not expected to have a page key
                target_page_num_reported = None
                p_index = link.get('page')
                
                if p_index is not None:
                    try:
                        # 1. Cast to int (handles the string/int confusion)
                        p_index_int = int(p_index)
                        
                        # 2. Logic Clamp: PyMuPDF sometimes reports the 'next' page 
                        # if a link points to the very bottom/edge of the target.
                        # If the index is >= total pages, clamp it to the last page.
                        if p_index_int >= doc.page_count:
                            p_index_int = doc.page_count - 1
                        
                        target_page_num_reported = p_index_int + 1
                    except (ValueError, TypeError):
                        target_page_num_reported = "Error"

                if link['kind'] == fitz.LINK_URI:
                    target =  link.get('uri', 'URI (Unknown Target)')
                    link_dict.update({
                        'type': 'External (URI)',
                        'url': link.get('uri'),
                        'target': target
                    })
                
                elif link['kind'] == fitz.LINK_GOTO:
                    target = f"Page {target_page_num_reported}"
                    link_dict.update({
                        'type': 'Internal (GoTo/Dest)',
                        'destination_page': target_page_num_reported,
                        'destination_view': destination_view,
                        'target': target
                    })
                
                elif link['kind'] == fitz.LINK_GOTOR:
                    link_dict.update({
                        'type': 'Remote (GoToR)',
                        'remote_file': link.get('file'),
                        'destination': destination_view
                    })
                
                elif link.get('page') is not None and link['kind'] != fitz.LINK_GOTO: 
                    target = f"Page {target_page_num_reported}"
                    link_dict.update({
                        'type': 'Internal (Resolved Action)',
                        'destination_page': target_page_num_reported,
                        'destination_view': destination_view,
                        'source_kind': link.get('kind'),
                        'target': target
                    })
                    
                else:
                    target = link.get('url') or link.get('remote_file') or link.get('target')
                    link_dict.update({
                        'type': 'Other Action',
                        'action_kind': link.get('kind'),
                        'target': target
                    })

                ## --- General Serialization Cleaner ---
                #for key, value in link_dict.items():
                #    if hasattr(value, 'rect') and hasattr(value, 'point'):
                #        # This handles Rect and Point objects that may slip through
                #        link_dict[key] = str(value)
                ## --- End Cleaner ---
                    
                links_data.append(link_dict)

        doc.close()
    except Exception as e:
        print(f"An error occurred: {e}", file=sys.stderr)
    return links_data


def call_stable():
    """
    Placeholder function for command-line execution (e.g., in __main__).
    Note: This requires defining PROJECT_NAME, CLI_MAIN_FILE, etc., or 
    passing them as arguments to run_report.
    """
    from pdflinkcheck.report import run_report_and_call_exports
    
    run_report_and_call_exports(pdf_library = "pymupdf")

if __name__ == "__main__":
    call_stable()
